/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: June 15th 2005                              */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include "arithmeticcoder.h"

const unsigned int g_FirstQuarter = 0x20000000;
const unsigned int g_ThirdQuarter = 0x60000000;
const unsigned int g_Half         = 0x40000000;

void initarithmeticcoder(ARCODER *ac)
{
	ac->mBitCount = 0;
	ac->mBitBuffer = 0;

	ac->mLow = 0;
	ac->mHigh = 0x7FFFFFFF; // just work with least significant 31 bits
	ac->mScale = 0;

	ac->mBuffer = 0;
	ac->mStep = 0;
}

void ACSetBuffer(BYTE *buffer, int *length,ARCODER *ac)
{
	ac->mByteBuffer = buffer;
	ac->ByteBufferPointer=0;
	ac->ByteBufferLength = length;
}

void ACSetBit( const unsigned char bit,ARCODER *ac )
{
	// add bit to the buffer
	ac->mBitBuffer = (ac->mBitBuffer << 1) | bit;
	ac->mBitCount++;

	if(ac->mBitCount == 8) // buffer full
	{
		// write
		ac->mByteBuffer[ac->ByteBufferPointer]=ac->mBitBuffer;
		ac->ByteBufferPointer++;
		ac->mBitCount = 0;
	}
}

void ACSetBitFlush(ARCODER *ac)
{
	// fill buffer with 0 up to the next byte
	while( ac->mBitCount != 0 )
		ACSetBit( 0,ac );
}

unsigned char ACGetBit(ARCODER *ac)
{
	unsigned char bit;
	if(ac->mBitCount == 0) // buffer empty
	{
		if( ac->ByteBufferPointer < *(ac->ByteBufferLength) )
		{ // file read completely?
			ac->mBitBuffer=ac->mByteBuffer[ac->ByteBufferPointer];
			ac->ByteBufferPointer++;
		}
		else
			ac->mBitBuffer = 0; // append zeros

		ac->mBitCount = 8;
	}

	// extract bit from buffer
	bit = ac->mBitBuffer >> 7;
	ac->mBitBuffer <<= 1;
	ac->mBitCount--;

	return bit;
}

void ACEncode( const unsigned int low_count,const unsigned int high_count,const unsigned int total,ARCODER *ac )
// total < 2^29
{
	// partition number space into single steps
	ac->mStep = ( ac->mHigh - ac->mLow + 1 ) / total; // interval open at the top => +1

	// update upper bound
	ac->mHigh = ac->mLow + ac->mStep * high_count - 1; // interval open at the top => -1

	// update lower bound
	ac->mLow = ac->mLow + ac->mStep * low_count;

	// apply e1/e2 mapping
	while( ( ac->mHigh < g_Half ) || ( ac->mLow >= g_Half ) )
		{
			if( ac->mHigh < g_Half )
			{
				ACSetBit( 0,ac );
				ac->mLow = ac->mLow * 2;
				ac->mHigh = ac->mHigh * 2 + 1;

				// perform e3 mappings
				for(; ac->mScale > 0; ac->mScale-- )
					ACSetBit( 1,ac );
			}
			else if( ac->mLow >= g_Half )
			{
				ACSetBit( 1,ac );
				ac->mLow = 2 * ( ac->mLow - g_Half );
				ac->mHigh = 2 * ( ac->mHigh - g_Half ) + 1;

				// perform e3 mappings
				for(; ac->mScale > 0; ac->mScale-- )
					ACSetBit( 0,ac );
			}
		}

	// e3
	while( ( g_FirstQuarter <= ac->mLow ) && ( ac->mHigh < g_ThirdQuarter ) )
	{
		// keep necessary e3 mappings in mind
		ac->mScale++;
		ac->mLow = 2 * ( ac->mLow - g_FirstQuarter );
		ac->mHigh = 2 * ( ac->mHigh - g_FirstQuarter ) + 1;
	}
}

void ACEncodeFinish(ARCODER *ac)
{
	// There are two possibilities of how mLow and mHigh can be distributed,
	// which means that two bits are enough to distinguish them.
	unsigned int i;
	if( ac->mLow < g_FirstQuarter ) // mLow < FirstQuarter < Half <= mHigh
	{
		ACSetBit( 0,ac );

		for( i=0; i<ac->mScale+1; i++ ) // perform e3-skaling
			ACSetBit(1,ac);
	}
	else // mLow < Half < ThirdQuarter <= mHigh
	{
		ACSetBit( 1,ac ); // zeros added automatically by the decoder; no need to send them
	}

	// empty the output buffer
	ACSetBitFlush(ac);
	*(ac->ByteBufferLength)=ac->ByteBufferPointer;
}

void ACDecodeStart(ARCODER *ac)
{
	// Fill buffer with bits from the input stream
	int i;
	for( i=0; i<31; i++ ) // just use the 31 least significant bits
		ac->mBuffer = ( ac->mBuffer << 1 ) | ACGetBit(ac);
}

unsigned int ACDecodeTarget(const unsigned int total,ARCODER *ac)
// total < 2^29
{
	// split number space into single steps
	ac->mStep = ( ac->mHigh - ac->mLow + 1 ) / total; // interval open at the top => +1

	// return current value
	return ( ac->mBuffer - ac->mLow ) / ac->mStep;
}

void ACDecode( const unsigned int low_count,const unsigned int high_count,ARCODER *ac )
{
	// update upper bound
	ac->mHigh = ac->mLow + ac->mStep * high_count - 1; // interval open at the top => -1

	// update lower bound
	ac->mLow = ac->mLow + ac->mStep * low_count;

	// e1/e2 mapping
	while( ( ac->mHigh < g_Half ) || ( ac->mLow >= g_Half ) )
		{
			if( ac->mHigh < g_Half )
			{
				ac->mLow = ac->mLow * 2;
				ac->mHigh = ac->mHigh * 2 + 1;
				ac->mBuffer = 2 * ac->mBuffer + ACGetBit(ac);
			}
			else if( ac->mLow >= g_Half )
			{
				ac->mLow = 2 * ( ac->mLow - g_Half );
				ac->mHigh = 2 * ( ac->mHigh - g_Half ) + 1;
				ac->mBuffer = 2 * ( ac->mBuffer - g_Half ) + ACGetBit(ac);
			}
			ac->mScale = 0;
		}

	// e3 mapping
	while( ( g_FirstQuarter <= ac->mLow ) && ( ac->mHigh < g_ThirdQuarter ) )
	{
		ac->mScale++;
		ac->mLow = 2 * ( ac->mLow - g_FirstQuarter );
		ac->mHigh = 2 * ( ac->mHigh - g_FirstQuarter ) + 1;
		ac->mBuffer = 2 * ( ac->mBuffer - g_FirstQuarter ) + ACGetBit(ac);
	}
}

/* Model*/
void ModelProcess(BYTE *source,BYTE *target,int Nsource,int *Ntarget,enum ModeE mode,MODEL *mmodel)
{
	mmodel->mSource = source;
	mmodel->mTarget = target;
	mmodel->nSource = Nsource;
	mmodel->mSourcePointer=0;
	mmodel->mTargetPointer=0;

	if( mode == MODE_ENCODE )
	{
		ACSetBuffer( mmodel->mTarget,Ntarget,&(mmodel->mAC) );

		// encode
		ModelEncode(mmodel);

		ACEncodeFinish(&(mmodel->mAC));
	}
	else // MODE_DECODE
	{
		ACSetBuffer( mmodel->mSource,&Nsource,&(mmodel->mAC));

		ACDecodeStart(&(mmodel->mAC));

		// decode
		ModelDecode(mmodel);
		*Ntarget=mmodel->mTargetPointer;
	}
};

void initmodel(MODEL *mmodel)
{
	// initialize probabilities with 1
	unsigned int i;
	initarithmeticcoder(&(mmodel->mAC));

	mmodel->mTotal = 257; // 256 + escape symbol for termination
	for( i=0; i<257; i++ )
		mmodel->mCumCount[i] = 1;
}

void ModelEncode(MODEL *mmodel)
{

	while( mmodel->mSourcePointer < mmodel->nSource )
	{
		unsigned char symbol;

		// read symbol
		symbol = mmodel->mSource[mmodel->mSourcePointer];
		mmodel->mSourcePointer++;

		if( mmodel->mSourcePointer <= mmodel->nSource )
		{
			// cumulate frequencies
			unsigned int low_count = 0;
			unsigned char j;
			for(  j=0; j<symbol; j++ )
				low_count += mmodel->mCumCount[j];

			// encode symbol
			ACEncode( low_count, low_count + mmodel->mCumCount[j], mmodel->mTotal,&(mmodel->mAC));

			// update model
			mmodel->mCumCount[ symbol ]++;
			mmodel->mTotal++;
		}
	}

	// write escape symbol for termination
	ACEncode( mmodel->mTotal-1, mmodel->mTotal, mmodel->mTotal,&(mmodel->mAC));
}

void ModelDecode(MODEL *mmodel)
{
	unsigned int symbol;

	do
	{
		unsigned int value;
		unsigned int low_count = 0;

		// read value
		value = ACDecodeTarget( mmodel->mTotal,&(mmodel->mAC) );


		// determine symbol
		for( symbol=0; low_count + mmodel->mCumCount[symbol] <= value; symbol++ )
			low_count += mmodel->mCumCount[symbol];

		// write symbol
		if( symbol < 256 )
		{
			mmodel->mTarget[mmodel->mTargetPointer]=symbol;
			mmodel->mTargetPointer++;
		}

		// adapt decoder
		ACDecode( low_count, low_count + mmodel->mCumCount[ symbol ],&(mmodel->mAC) );

		// update model
		mmodel->mCumCount[ symbol ]++;
		mmodel->mTotal++;
	}
	while( symbol != 256 ); // until termination symbol read
}

int arithmeticencode(BYTE *input,BYTE *output,int Ninput)
{
	MODEL mmodel;
	int Noutput;

	// choose model, here just order-0
	initmodel(&mmodel);

	/* Encode */
	//cout << "Encoding " << endl;
	ModelProcess( input, output, Ninput, &Noutput,MODE_ENCODE,&mmodel);

	return Noutput;
}

int arithmeticdecode(BYTE *input,BYTE *output,int Ninput)
{
	MODEL mmodel;
	int Noutput;

	// choose model, here just order-0
	initmodel(&mmodel);

	//cout << "Decoding " << endl;
	ModelProcess( input, output, Ninput, &Noutput,MODE_DECODE,&mmodel);


	return Noutput;
}