/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: June 8th 2005                               */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "wav.h"


/* Reads the information of a WAV file pointed by wav.
	Returns:
		 0: On success.
		-1: If wav is not a WAV file.
		-2: If the WAV is not in PCM format.
		-3: If the file is not PCM 44100Hz, 16bits, stereo.
	For 0 and -3, the arguments samplerate, bitspersample, channels and nsamples contain
	the sampling frequency, bits per sample, number of channels and the number of samples
	in the file. */
int readinfo(WAVFILE *wav)
{
	struct riff_hdr riffheader;
	struct wav_hdr wavheader;
	struct wavfmtck format;
	struct datachunk_hdr dthdr;
	fseek(wav->fwav,0,SEEK_SET);
	fread(&riffheader,sizeof(struct riff_hdr),1,wav->fwav);
	fread(&wavheader,sizeof(struct wav_hdr),1,wav->fwav);
	fread(&format,sizeof(struct wavfmtck),1,wav->fwav);
	
	if(strncmp(wavheader.wave_id,"WAVE",4))
		return -1;
	if(format.wFormatTag!=WAVE_FORMAT_PCM)
		return -2;

	wav->samplerate=format.dwSamplesPerSec;
	wav->bitspersample=format.wBitsPerSample;
	wav->channels=format.wChannels;

	if(wav->channels!=2 || wav->samplerate!=44100 || wav->bitspersample!=16)
		return -3;

	/* Moves until it reaches the "data" chunk which contains the samples */
	dthdr.chunkSize=wavheader.len-sizeof(struct wavfmtck);
	do
	{
		fseek(wav->fwav,dthdr.chunkSize,SEEK_CUR);
		fread(&dthdr,sizeof(struct datachunk_hdr),1,wav->fwav);
	}while(strncmp(dthdr.chunkID,"data",4));

	/* Computes number of samples */
	wav->totalsamples = dthdr.chunkSize / wav->channels / (wav->bitspersample/8);
	wav->samplesread=0;

	return 0;
}

/* Reads the next sample of the wav file and stores it in sample.
	Returns:
		 0: On success.
		-1: On failure (most likely end of file).
	WARNING: This function only reads the data located at the current file position.
	You should use readinfo before using this function for the first time. */
int readsample(WAVFILE *wav,struct sample *ssample)
{
	int i;
	if(wav->samplesread < wav->totalsamples)
		i=fread(ssample,sizeof(struct sample),1,wav->fwav);
	if(i!=1)
		return -1;
	wav->samplesread++;
	return 0;
}

void writeinfo(WAVFILE *wav)
{
	struct riff_hdr riffheader;
	struct wav_hdr wavheader;
	struct wavfmtck format;
	struct datachunk_hdr dthdr;
	fseek(wav->fwav,0,SEEK_SET);

	wav->samplesread=0;
	
	riffheader.id[0]='R';
	riffheader.id[1]='I';
	riffheader.id[2]='F';
	riffheader.id[3]='F';
	riffheader.len=wav->totalsamples*wav->bitspersample/8*wav->channels + sizeof(struct wav_hdr) + 
		sizeof(struct wavfmtck) + sizeof(struct datachunk_hdr);

	wavheader.wave_id[0]='W';
	wavheader.wave_id[1]='A';
	wavheader.wave_id[2]='V';
	wavheader.wave_id[3]='E';
	wavheader.id[0]='f';
	wavheader.id[1]='m';
	wavheader.id[2]='t';
	wavheader.id[3]=' ';
	wavheader.len=sizeof(struct wavfmtck);

	format.wFormatTag=WAVE_FORMAT_PCM;
	format.dwSamplesPerSec=wav->samplerate;
	format.wBitsPerSample=wav->bitspersample;
	format.wChannels=wav->channels;
	format.wBlockAlign=4;
	format.dwAvgBytesPerSec=wav->bitspersample/8 * wav->samplerate*wav->channels;

	dthdr.chunkID[0]='d';
	dthdr.chunkID[1]='a';
	dthdr.chunkID[2]='t';
	dthdr.chunkID[3]='a';
	dthdr.chunkSize = wav->totalsamples * wav->bitspersample/8 * wav->channels;

	fwrite(&riffheader,sizeof(struct riff_hdr),1,wav->fwav);
	fwrite(&wavheader,sizeof(struct wav_hdr),1,wav->fwav);
	fwrite(&format,sizeof(struct wavfmtck),1,wav->fwav);
	fwrite(&dthdr,sizeof(struct datachunk_hdr),1,wav->fwav);

}

int writesample(WAVFILE *wav,double left,double right)
{
	int i;
	struct sample samplewav;
	double norm;
	if(wav->samplesread < wav->totalsamples)
	{
		norm=fabs(left);
		if(norm<1.0)
			norm=1.0;
		samplewav.wLeft=(WORD)(left*32767.0/norm);
		norm=fabs(right);
		if(norm<1.0)
			norm=1.0;
		samplewav.wRight=(WORD)(right*32767.0/norm);
		i=fwrite(&samplewav,sizeof(struct sample),1,wav->fwav);
	}
	if(i!=1)
		return -1;
	wav->samplesread++;
	return 0;
}

int writewav(WAVFILE *wav,double *left,double *right,double maxl,double maxr)
{
	unsigned int i;
	struct sample samplewav;

	/* Signal is within range, no need to adjust the dynamic range */
	if(maxl<1.0)
		maxl=1.0;
	if(maxr<1.0)
		maxr=1.0;

	writeinfo(wav);
	for(i=0;i<wav->totalsamples;i++)
	{
		samplewav.wLeft=(WORD)(left[i]*32767.0/maxl);
		samplewav.wRight=(WORD)(right[i]*32767.0/maxr);
		fwrite(&samplewav,sizeof(struct sample),1,wav->fwav);
	}

	return 0;
}

/* Returns:
		0 on success
		same code as readinfo otherwise. 
	Allocates the needed memory to *left and *right */
int readwav(WAVFILE *wav,double **left,double **right)
{
	struct sample ssample;
	unsigned int count;
	int ret;

	ret=readinfo(wav);
	if(ret)
		return ret;

	count=0;
	*left=(double *)malloc(sizeof(double)*(wav->totalsamples));
	*right=(double *)malloc(sizeof(double)*(wav->totalsamples));
	while(!readsample(wav,&ssample))
	{
		//printf("%d || %1.8f %1.8f\n",cuenta,muestra.wLeft/32768.0,muestra.wRight/32768.0);
		(*left)[count]=ssample.wLeft/32768.0;
		(*right)[count]=ssample.wRight/32768.0;
		count++;
	}
	return 0;
}
